"""
Makes use of the FLB_MAX_FAILURES and FLB_BLOCK_INTERVAL values in
settings.py. If these are not present, default values of 5 failures
and 1 day are used.

The easiest way to remove a block is to delete the FailedAttempt record (e.g.
via the admin).
"""
from datetime import timedelta

from django.db import models
from django.utils import timezone
from django.conf import settings

# default values that can be overriden in settings.py
FLB_MAX_FAILURES = int(getattr(settings, 'FLB_MAX_FAILURES', 5))
FLB_BLOCK_INTERVAL = int(getattr(settings, 'FLB_BLOCK_INTERVAL', 1440))


class FailedAttempt(models.Model):
    username = models.CharField('Username', max_length=255)
    failures = models.PositiveIntegerField('Failures', default=0)
    timestamp = models.DateTimeField('Last failed attempt', auto_now=True)

    def too_many_failures(self):
        """
        Check if the minimum number of failures needed for a block
        has been reached
        """
        return self.failures >= FLB_MAX_FAILURES

    def recent_failure(self):
        """
        Checks if the timestamp one the FailedAttempt object is
        recent enough to result in an increase in failures
        """
        return timezone.now() < self.timestamp + timedelta(minutes=FLB_BLOCK_INTERVAL)

    def blocked(self):
        """
        Shortcut function for checking both too_many_failures
        and recent_failure
        """
        return self.too_many_failures() and self.recent_failure()
    blocked.boolean = True

    def __unicode__(self):
        return u'%s (%d failures until %s): ' % \
               (self.username, self.failures, self.timestamp)

    class Meta:
        ordering = ['-timestamp']



"""Admin settings for the failedloginblocker module"""

from django.contrib import admin

from .models import FailedAttempt


class AdminFailedAttempt(admin.ModelAdmin):
    list_display = ('username', 'failures', 'timestamp', 'blocked')
    search_fields = ('username',)

admin.site.register(FailedAttempt, AdminFailedAttempt)



from django.conf import settings

class LoginBlockedError(Exception):
    def __init__(self):
        msg = "Your account has been locked due to too many failed login attempts."
        msg += " Contact us to have your account reactivated."
        msg += " Alternatively wait {0} minutes then try again.".format(settings.FLB_BLOCK_INTERVAL)
        super(LoginBlockedError, self).__init__(msg)



""" Decorators used by failedloginblocker """

from .models import FailedAttempt
from .exceptions import LoginBlockedError


def monitor_login(auth_func):
    """
    Function that decorates Django authenticate() function with one that
    tracks failed logins and blocks further attempts based on a threshold
    """

    if hasattr(auth_func, '__PROTECT_FAILED_LOGINS__'):
        # avoiding multiple decorations
        return auth_func

    def decorate(*args, **kwargs):
        """ Wrapper for Django authenticate function """
        user = kwargs.get('username', '')
        if not user:
            raise ValueError('username must be supplied by the \
                authentication function for FailedLoginBlocker to operate')

        try:
            fa = FailedAttempt.objects.get(username=user)
            if fa.recent_failure():
                if fa.too_many_failures():
                    raise LoginBlockedError()
            else:
                # the block interval is over, reset the count
                fa.failures = 0
                fa.save()
        except FailedAttempt.DoesNotExist:
            fa = None

        result = auth_func(*args, **kwargs)
        if result:
            # the authentication was successful
            return result
        # authentication failed
        fa = fa or FailedAttempt(username=user, failures=0)
        fa.failures += 1
        fa.save()
        # return with unsuccessful auth
        return None

    decorate.__PROTECT_FAILED_LOGINS__ = True
    return decorate



from django.http import HttpResponse
from django.template import Context, loader

from .exceptions import LoginBlockedError


class FailedLoginBlockerMiddleware(object):
    """
    When a `LoginBlockedError` is raised this middleware intercepts it and renders
    a blocked template with a 403 status.
    """

    def process_exception(self, request, exception):
        if type(exception) is not LoginBlockedError:
            return

        template = loader.get_template('failedloginblocker/blocked.html')
        context = Context({
            'exception': exception,
        })
        response = HttpResponse(template.render(context))
        response.status_code = 403
        return response



from django.test import TestCase
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from django.contrib import auth

from .models import FailedAttempt
from .exceptions import LoginBlockedError


class FailedLoginBlockerTest(TestCase):
    def setUp(self):
        self.test_user = User.objects.create_user('test_user', 'test@example.com', 'password')
        self.login_url = reverse('login')

    def test_blocked_after_three_failed_login_attempts(self):
        credentials = {
            'username': self.test_user.username,
            'password': 'wrong_password',
        }

        self.client.login(**credentials)
        failed_attempt = FailedAttempt.objects.all()[0]
        self.assertEqual(self.test_user.username, failed_attempt.username)
        self.assertEqual(1, failed_attempt.failures)

        self.client.login(**credentials)
        failed_attempt = FailedAttempt.objects.all()[0]
        self.assertEqual(2, failed_attempt.failures)

        self.client.login(**credentials)
        failed_attempt = FailedAttempt.objects.all()[0]
        self.assertEqual(3, failed_attempt.failures)

        # After 3 failed login attempts, an exception should be raised.
        self.assertRaises(LoginBlockedError, self.client.login, **credentials)



from setuptools import setup

setup(
    name='django-failedloginblocker',
    version='1.0.1',
    packages=['failedloginblocker'],
    package_data={
        '': ['*.html'],
    },
    license='BSD (3-Clause) License',
    long_description=open('README').read(),
    maintainer='mySociety',
    maintainer_email='programmers@mysociety.org',
    url='https://github.com/mysociety/django-failedloginblocker',
)



from django.contrib.auth.backends import ModelBackend

from .decorators import monitor_login


class MonitoredModelBackend(ModelBackend):
    """
    Custom auth backend which monitors login attempts and blocks after
    a preconfigured number of failures.
    """

    @monitor_login
    def authenticate(self, **credentials):
        return super(MonitoredModelBackend, self).authenticate(**credentials)
